!--------------------------------------------------------------------------------
program InvInterp3D
  use sep
	use solver_reg_mod
	use cgstep_mod
	use igrad3
	use bin3
	use lint3
	use ndipfilt3
	use nhelix
	use morgan_util

  implicit none

!--------------------------------------------------------------------------------
  integer :: nd, m1, m2, m3, niter, stat, itest, iverb, ilint, ilev, j0, j1,    &
	           idip, iinv, inorm, imod, m(3)
  real    :: o1, d1, o2, d2, o3, d3, eps, d(3), o(3), eps1, eps2, eps3
	logical :: verb1, verb2, lint, test, dip, inv, norm, modeling

  real, dimension (:),   pointer :: mm, data, data_od, dip1, dip2, resd, resd_ms
  real, dimension (:,:), pointer :: xyz

	type(nfilter)      :: nfilt2, nfilt3
	logical, parameter :: T=.true., F=.false.

  call sep_init(SOURCE)

!--------------------------------------------------------------------------------
	call from_param( "niter", niter, 10);
	call from_param( "eps", eps, 1.0);
	call from_param( "eps1", eps1, 1.0);
	call from_param( "eps2", eps2, 1.0);
	call from_param( "eps3", eps3, 1.0);

	call from_param( "verb", iverb, 0);  verb1=(iverb==1);  verb2=(iverb==2)
	call from_param( "lint", ilint, 1);  lint=(ilint==1)
	call from_param( "test", itest, 0);  test=(itest==1)
	call from_param( "dip",  idip,  0);  dip=(idip==1)
	call from_param( "inv",  iinv,  1);  inv=(iinv==1)
	call from_param( "norm", inorm, 0);  norm=(inorm==1)

	call from_param( "modeling", imod, 0);  modeling=(imod==1)

	!------------------------------------------------------------------------------
	if( .NOT. modeling) then                   ! Output is gridded, input irregular
	  call from_param( "n1", m1)
		call from_param( "n2", m2)
		call from_param( "n3", m3)
		call from_param( "o1", o1, 0.0)
		call from_param( "o2", o2, 0.0)
		call from_param( "o3", o3, 0.0)
		call from_param( "d1", d1, 1.0)
		call from_param( "d2", d2, 1.0)
		call from_param( "d3", d3, 1.0)
  	call from_history("n1", nd)
  	call to_history( "n1", m1)
	  call to_history( "n2", m2)
	  call to_history( "n3", m3)
		call to_history( "o1", o1)
		call to_history( "o2", o2)
		call to_history( "o3", o3)
		call to_history( "d1", d1)
		call to_history( "d2", d2)
		call to_history( "d3", d3)
	else                                       ! Output is irregular, input gridded
	  call from_history( "n1", m1 )
		call from_history( "n2", m2 )
		call from_history( "n3", m3 )  
		call from_history( "o1", o1 )
		call from_history( "o2", o2 )
		call from_history( "o3", o3 )
		call from_history( "d1", d1 )
		call from_history( "d2", d2 )
		call from_history( "d3", d3 )
		call from_aux( "xyz", "n1", nd)
		call to_history( "n1", nd)
		call to_history( "n2", 1)
		call to_history( "n3", 1)
	end if
	!------------------------------------------------------------------------------

	m=(/m1,m2,m3/);  d=(/d1,d2,d3/);  o=(/o1,o2,o3/)

	allocate( mm(m1*m2*m3), data(nd), xyz(nd,3) )

	if( .NOT. modeling) then
		call sep_read( data)
		call sep_read( xyz )
		if(exist_file("m0")) call sep_read( mm, "m0")
	else
		call sep_read( mm )
		call sep_read( xyz, "xyz" )
	end if

	call igrad3_init( m1, m2, m3, eps1=eps1, eps2=eps2, eps3=eps3)


	if(test) then
		if(lint) then
			call lint3_test()
		else
			call bin3_test()
		end if
	end if

	if( inv ) then                                                 ! CG inversion
		allocate( resd(nd) )
		if( dip) then
			allocate( dip1(m1*m2*m3), dip2(m1*m2*m3) )
			dip1 = 0.0;  if(exist_file("dip1")) call sep_read( dip1, "dip1")
			dip2 = 0.0;  if(exist_file("dip2")) call sep_read( dip2, "dip2")
			nfilt2 = get_ndipfilt3( dip1, m, d, 9, 2)
			nfilt3 = get_ndipfilt3( dip2, m, d, 9, 3)
			call ndipfilt3_init( nfilt2, nfilt3, m )
		
			if( exist_file("impresp") ) then              ! NS Dip Filter Inverse IR
				call to_history("n1", m1, "impresp")
				call to_history("n2", m2, "impresp")
				call to_history("n3", m3, "impresp")
				call to_history("n4", 2, "impresp")
				call ndipfilt3_test( nfilt2, (/m1,m2,m3/), mm, 2)       ! Warning: m0
				call sep_write(mm, "impresp")                           ! overwritten 
				call ndipfilt3_test( nfilt3, (/m1,m2,m3/), mm, 3)       ! Warning: m0
				call sep_write(mm, "impresp")                           ! overwritten 
				mm = 0.0
			end if
		end if

		if( lint) then                                    ! Trilinear interpolation
			call lint3_init( m, o, d, xyz, norm=norm, verb=verb2 )

			if( dip) then
				call solver_reg( mm, data, lint3_exec, ndipfilt23_exec, cgstep, 2*m1*m2*m3,  &
				                 niter, eps, verb=verb1, m0=mm, resd=resd               )
				call ndipfilt3_close()
			else
				call solver_reg( mm, data, lint3_exec, igrad3_weighted_exec, cgstep, 3*m1*m2*m3,  &
				                 niter, eps, verb=verb1, m0=mm, resd=resd               )
			end if
		else                                         ! 3-D nearest-neighbor binning
			call bin3_init( m, o, d, xyz, norm=norm, verb=verb2 )
			
			if( dip) then
				call solver_reg( mm, data, bin3_exec, ndipfilt23_exec, cgstep, 2*m1*m2*m3,   &
				                 niter, eps, verb=verb1, m0=mm, resd=resd               )
				call ndipfilt3_close()
			else
				call solver_reg( mm, data, bin3_exec, igrad3_weighted_exec, cgstep, 3*m1*m2*m3,   &
				                 niter, eps, verb=verb1, m0=mm, resd=resd               )
			end if
		end if
	else                                         ! Just apply operator or adjoint
		if( lint) then                                    ! Trilinear interpolation
			call lint3_init( m, o, d, xyz, norm=norm, verb=verb2 )
			stat = lint3_exec( (.NOT. modeling), F, mm, data)
		else                                         ! 3-D nearest-neighbor binning
			call bin3_init( m, o, d, xyz, norm=norm, verb=verb2 )
			stat = bin3_exec( (.NOT. modeling), F, mm, data)
		end if
	end if

	if( .NOT. modeling) then
		call sep_write(mm)
		if( inv .AND. exist_file("resd") ) then
			call to_history( "n1", nd, "resd")
			call to_history( "n2", 1,  "resd")
			call to_history( "n3", 1,  "resd")
			call sep_write( resd)
			deallocate( resd)
		end if
	else
		call sep_write(data)
	end if

  deallocate( mm, data, xyz)

  call exit(0) 

end program InvInterp3D
!-------------------------------------------------------------------------------
