!--------------------------------------------------------------------------------
program Losmooth3
  use sep
!	use mult_nmo_util

  implicit none

  integer :: nd, id, n1, n2, n3, i1, i2, i3, i, ndone, kmin1, kmin2, kmin3, k1, &
	           k2, k3, j1_0, j1_1, j2_0, j2_1, j3_0, j3_1, outwt, sign1, sign2,   &
	           sign3, nbinfull, maxlev, nlev
	real    :: d1, d2, d3, o1, o2, o3, dmin1, dmin2, dmin3, dist1, dist2, dist3,  &
	           x1, x2, x3, dsum, pow, dthresh, expand1, expand2, expand3
	logical :: allbinsfull

  real, dimension (:),     pointer :: data
  real, dimension (:,:),   pointer :: txy
  real, dimension (:,:,:), pointer :: model, wtsum
  integer, dimension (:,:,:), pointer :: binct

  call sep_init(SOURCE)

  call from_history("n1", nd)                          ! Read irregular pick data

	call from_param( "n1", n1 )
	call from_param( "n2", n2 )
	call from_param( "n3", n3 )
	call from_param( "o1", o1, 0.0)
	call from_param( "o2", o2, 0.0)
	call from_param( "o3", o3, 0.0)
	call from_param( "d1", d1, 0.004)
	call from_param( "d2", d2, 1.0)
	call from_param( "d3", d3, 1.0)
	call from_param( "dmin1", dmin1, 5*d1)
	call from_param( "dmin2", dmin2, 5*d2)
	call from_param( "dmin3", dmin3, 5*d3)
	call from_param( "expand1", expand1, 1.2)
	call from_param( "expand2", expand2, 1.2)
	call from_param( "expand3", expand3, 1.2)
	call from_param( "outwt", outwt, 0)
	call from_param( "dropoff", pow, 1.0)
	call from_param( "nbinfull", nbinfull, 4)
	call from_param( "preserve", dthresh, 1.0)
	call from_param( "maxlev", maxlev, 20)

	if( nd < nbinfull) call seperr("ERROR: require # data > nbinfull")

	allocate( data(nd), txy(nd,3), model(n1,n2,n3), wtsum(n1,n2,n3),              &
	          binct(n1,n2,n3)                                                     )
	call sep_read( data)
	call sep_read( txy)

	kmin1 = nint( dmin1/d1)         ! Integer radii
	kmin2 = nint( dmin2/d2)
	kmin3 = nint( dmin3/d3)

	allbinsfull=.false.
	binct = 0
	nlev = 0
	sign1 = isign( 1.0, d1)
	sign2 = isign( 1.0, d2)
	sign3 = isign( 1.0, d3)

	do while( .NOT. allbinsfull)

		ndone = 0
		do id = 1,nd
			k1 = nint( (txy(id,1) - o1)/d1 + 1)
			k2 = nint( (txy(id,2) - o2)/d2 + 1)
			k3 = nint( (txy(id,3) - o3)/d3 + 1)

			j1_0 = max( 1, min(n1,k1-kmin1*sign1) )
			j1_1 = max( 1, min(n1,k1+kmin1*sign1) )
			j2_0 = max( 1, min(n2,k2-kmin2*sign2) )
			j2_1 = max( 1, min(n2,k2+kmin2*sign2) )
			j3_0 = max( 1, min(n3,k3-kmin3*sign3) )
			j3_1 = max( 1, min(n3,k3+kmin3*sign3) )
	
			do i3=j3_0,j3_1
				do i2=j2_0,j2_1
					do i1=j1_0,j1_1
						if( binct(i1,i2,i3) < nbinfull ) then     ! Only work on nonfull bins
							x1 = (i1-1)*d1 + o1
							x2 = (i2-1)*d2 + o2
							x3 = (i3-1)*d3 + o3
							dist1 = ( (x1-txy(id,1))/d1 )**2
							dist2 = ( (x2-txy(id,2))/d2 )**2
							dist3 = ( (x3-txy(id,3))/d3 )**2
							dsum = sqrt( dist1 + dist2 + dist3)
							if( dsum > dthresh ) then
								model(i1,i2,i3) = model(i1,i2,i3) + data(id)/(dsum**pow)
								wtsum(i1,i2,i3) = wtsum(i1,i2,i3) + 1.0/(dsum**pow)
							else
								model(i1,i2,i3) = model(i1,i2,i3) + data(id)/(dthresh**pow)
								wtsum(i1,i2,i3) = wtsum(i1,i2,i3) + 1/(dthresh**pow)
							end if
							binct(i1,i2,i3) = binct(i1,i2,i3) + 1
						end if
					end do
				end do
			end do

			if( mod(ndone,nd/10)==0 ) write(0,"(a,i,a,i)") 'Done ',ndone,' of',nd
			ndone = ndone+1
		end do

		if( any( binct < nbinfull) .AND. nlev<maxlev ) then     ! still working
			kmin1 = ceiling(sign1*kmin1*expand1)
			kmin2 = ceiling(sign1*kmin2*expand2)
			kmin3 = ceiling(sign1*kmin3*expand3)
			nlev = nlev + 1
			write(0,"(a,i3,a,i)") 'Level',nlev,': # Non-full bins: ',count(binct<nbinfull)
			write(0,"(a,3i4)") 'Expanding window to ',kmin1,kmin2,kmin3
		else
			allbinsfull = .true.
		end if

	end do

	where(wtsum>0.0) model = model/wtsum

	call to_history( "n1", n1)
	call to_history( "n2", n2)
	call to_history( "n3", n3)
	call to_history( "o1", o1)
	call to_history( "o2", o2)
	call to_history( "o3", o3)
	call to_history( "d1", d1)
	call to_history( "d2", d2)
	call to_history( "d3", d3)

	call sep_write( model)

	if(outwt==1) then
		call to_history( "n4", 2)
		call sep_write( wtsum)
	end if

  deallocate( txy, data, model, wtsum, binct )

  call exit(0) 

end program Losmooth3
!--------------------------------------------------------------------------------
