!-------------------------------------------------------------------------------
program Random_known

  use sep
	use cartesian

  implicit none

  real	  :: d1, d2, d3, o1, o2, o3, f, x1, x2, x3
  integer :: n1, n2, n3, npts, ipts, i1, i2, i3, seed, itrc, j, irand, ireg, nbad
	logical :: trace, rand
  real, dimension(:,:,:), pointer  :: model, fold
  real, dimension(:,:), pointer    :: data, xyz

  call sep_init(SOURCE)
  call from_history( "n1", n1)
	call from_history( "n2", n2)
	call from_history( "n3", n3)
  call from_history( "o1", o1)
  call from_history( "o2", o2)
  call from_history( "o3", o3)
	call from_history( "d1", d1)
	call from_history( "d2", d2)
	call from_history( "d3", d3)
	call from_param( "npts", npts, n1*n2*n3/100 )
	call from_param( "seed", seed, 884239805 )
	call from_param( "trace", itrc, 1);  trace=(itrc==1)
	call from_param( "rand", irand, 1);  rand=(irand==1)

	if( .NOT. rand) then
		if(trace) then
			npts = n2*n3
		else
			npts = n1*n2*n3
		end if
	end if

  allocate( model(n1,n2,n3), fold(n1,n2,n3) )
	if( trace) then
		call to_history( "n1", n1)
		call to_history( "n2", npts)
		call to_history( "n3", 1)
		call to_history( "n1", npts,"xyz")
		call to_history( "n2", 2,"xyz")
		call to_history( "n3", 1,"xyz")
	  allocate( data(n1,npts), xyz(npts,2) )
	else
		call to_history( "n1", npts)
		call to_history( "n2", 1)
		call to_history( "n3", 1)
		call to_history( "n1", npts,"xyz")
		call to_history( "n2", 3,"xyz")
		call to_history( "n3", 1,"xyz")
	  allocate( data(npts,1), xyz(npts,3) )
	end if

	call sep_read( model)

	call RANDOM_SEED( size=seed )
	fold = 0.0
	
	ipts = 1
	nbad = 0
	if( rand) then                                 ! Random sampling
		do while( ipts <= npts)
			call RANDOM_NUMBER( f)
			i2 = min(n2,max(1,floor(abs(n2*f))))
			x2 = (i2-1)*d2 + o2
			call RANDOM_NUMBER( f)
			i3 = min(n3,max(1,floor(abs(n3*f))))
			x3 = (i3-1)*d3 + o3
			if( trace) then           ! Don't subsample 1-axis if this is "seismic" data
				if( all(fold(:,i2,i3)==0.0) ) then        ! Haven't done this bin already
					do i1=1,n1
						data(i1,ipts) = model(i1,i2,i3)
						xyz(ipts,1) = x2
						xyz(ipts,2) = x3
						fold(i1,i2,i3) = 1.0
					end do
					ipts = ipts + 1
				else
					nbad = nbad + 1
					if( nbad > 100) then
						write(0,*) 'old seed:',seed
						seed = seed+107-nbad
						write(0,*) 'new seed:',seed
						call RANDOM_SEED( size=seed )
						nbad = 0
					end if
				end if
			else
				call RANDOM_NUMBER( f)
				i1 = min(n1,max(1,floor(abs(n1*f))))
				if( fold(i1,i2,i3)==0.0 ) then           ! Haven't done this bin already
					x1 = (i1-1)*d1 + o1
					data(ipts,1) = model(i1,i2,i3)
					xyz(ipts,1) = x1
					xyz(ipts,2) = x2
					xyz(ipts,3) = x3
					fold(i1,i2,i3) = 1.0
					ipts = ipts + 1
				else
					nbad = nbad + 1
					if( nbad > 100) then
						write(0,*) 'old seed:',seed
						seed = seed+107-nbad
						write(0,*) 'new seed:',seed
						call RANDOM_SEED( size=seed )
						nbad = 0
					end if
				end if
			end if
		end do
	else                                            ! Just copy from grid to (t,x,y)
		do i3=1,n3
			x3 = (i3-1)*d3 + o3
			do i2=1,n2
				x2 = (i2-1)*d2 + o2
				do i1=1,n1
					if( trace) then
						data(i1,ipts) = model(i1,i2,i3)
						xyz(ipts,1) = x2
						xyz(ipts,2) = x3
					else
						x1 = (i1-1)*d1 + o1
						data(ipts,1) = model(i1,i2,i3)
						xyz(ipts,1) = x1
						xyz(ipts,2) = x2
						xyz(ipts,3) = x3
						ipts = ipts + 1	
					end if
					fold(i1,i2,i3) = 1.0
				end do
				if(trace) ipts=ipts+1
			end do
		end do
	end if

  call sep_write( data)
  call sep_write( xyz, "xyz")

	if( exist_file("fold")) then
		call to_history("n1",n1,"fold")
		call to_history("n2",n2,"fold")
		call to_history("n3",n3,"fold")
		call to_history("d1",d1,"fold")
		call to_history("d2",d2,"fold")
		call to_history("d3",d3,"fold")
		call to_history("o1",o1,"fold")
		call to_history("o2",o2,"fold")
		call to_history("o3",o3,"fold")
		call sep_write(fold,"fold")
	end if

	deallocate( fold, model, data, xyz)

  call exit(0)

end program Random_known
!-------------------------------------------------------------------------------
