!--------------------------------------------------------------------------------
! 3-D gradient with adjoint,  r= grad( p)
!--------------------------------------------------------------------------------
module igrad3  

  use adj_mod
	use chain0_mod
	use scaled_identity
	use dottest

  implicit none

  integer, private, save :: n1, n2, n3
	real, private, save    :: eps(3)
	logical, private, save, parameter :: T = .true., F = .false.
	real, dimension(:), pointer, private, save :: tmp

  contains

!--------------------------------------------------------------------------------
  subroutine igrad3_init( n1_in, n2_in, n3_in, eps1, eps2, eps3 )
    integer    :: n1_in, n2_in, n3_in 
		real       :: eps1, eps2, eps3
		optional   :: eps1, eps2, eps3

    n1 = n1_in;  n2 = n2_in;  n3 = n3_in
		allocate( tmp(n1*n2*n3*3) )

		eps(1)=1.0; if(present(eps1)) eps(1)=eps1
		eps(2)=1.0; if(present(eps2)) eps(2)=eps2
		eps(3)=1.0; if(present(eps3)) eps(3)=eps3
  end subroutine igrad3_init
!--------------------------------------------------------------------------------

!--------------------------------------------------------------------------------
  function igrad3_lop ( adj, add, p, r) result(stat)
    integer            :: stat 
    logical,intent(in) :: adj,add 
    real,dimension(:)  :: p,r 
    call adjnull (adj,add,p,r )
		if( n3==1 ) then
	    call igrad3_2d_lop2(adj,add,p,r )
		else
	    call igrad3_lop2(adj,add,p,r )
		end if
    stat=0
  end function 
!--------------------------------------------------------------------------------

!--------------------------------------------------------------------------------
  subroutine igrad3_2d_lop2(adj,add,p,r)
    logical,intent(in) :: adj,add 
    real, dimension (n1,n2)  :: p 
    real, dimension (n1,n2,3)  :: r 
    integer i,j,ipl,jpl
    do i= 1, max(1,n1-1)
			ipl = i+1
			if(n1==1) ipl = 1
      do j= 1, max(1,n2-1)
				jpl = j+1
				if(n2==1) jpl = 1
        if ( adj) then
          p(ipl,j  )  =	p(ipl,j  )  +  r(i,j,1) 
          p(i  ,j  )  =	p(i  ,j  )  -  r(i,j,1) 
          p(i  ,jpl)  =	p(i  ,jpl)  +  r(i,j,2) 
          p(i  ,j  )  =	p(i  ,j  )  -  r(i,j,2)
        else
          r(i,j,1)  =	r(i,j,1)  +  ( p(ipl,j) - p(i,j) ) 
          r(i,j,2)  =	r(i,j,2)  +  ( p(i,jpl) - p(i,j) ) 
          r(i,j,3)  =	r(i,j,3)
        end if
      end do
    end do
  end subroutine 
!--------------------------------------------------------------------------------

!--------------------------------------------------------------------------------
  subroutine igrad3_lop2(adj,add,p,r)
    logical,intent(in) :: adj,add 
    real, dimension (n1,n2,n3)  :: p 
    real, dimension (n1,n2,n3,3)  :: r 
    integer i,j,k
    do i= 1, max(1,n1-1)
      do j= 1, max(1,n2-1)
        do k= 1, max(1,n3-1)
          if ( adj) then
            p(i+1,j  ,k  )  =					p(i+1,j  ,k  )  +  r(i,j,k,1) 
            p(i  ,j  ,k  )  =					p(i  ,j  ,k  )  -  r(i,j,k,1) 
            p(i  ,j+1,k  )  =					p(i  ,j+1,k  )  +  r(i,j,k,2) 
            p(i  ,j  ,k  )  =					p(i  ,j  ,k  )  -  r(i,j,k,2)
            p(i  ,j  ,k+1)  =					p(i  ,j  ,k+1)  +  r(i,j,k,3) 
            p(i  ,j  ,k  )  =					p(i  ,j  ,k  )  -  r(i,j,k,3)
          else
            r(i,j,k,1)  =					r(i,j,k,1)  +  ( p(i+1,j,k) - p(i,j,k)) 
            r(i,j,k,2)  =					r(i,j,k,2)  +  ( p(i,j+1,k) - p(i,j,k)) 
            r(i,j,k,3)  =					r(i,j,k,3)  +  ( p(i,j,k+1) - p(i,j,k))
          end if
        end do
      end do
    end do
  end subroutine 
!--------------------------------------------------------------------------------

!--------------------------------------------------------------------------------
  subroutine igrad3_damping_init(eps_damp)
		real, intent(IN) :: eps_damp
		call scaled_identity_init(eps_damp)
  end subroutine 
!--------------------------------------------------------------------------------

!--------------------------------------------------------------------------------
integer function igrad3_damping_exec( adj, add, model, data)
	real, dimension(:) :: model, data
	logical            :: adj, add
	integer            :: m0, m1
	
	m0 = 3*n1*n2*n3
	m1 = 4*n1*n2*n3

	call array0( igrad3_lop, scaled_identity_exec, adj, add, model, data(1:m0),   &
	             data(m0+1:m1)                                                    )

	igrad3_damping_exec = 0

end function igrad3_damping_exec
!--------------------------------------------------------------------------------

!--------------------------------------------------------------------------------
integer function igrad3_weighted_exec( adj, add, model, data)
	real, dimension(:) :: model, data
	logical            :: adj, add
	integer            :: m0, m1, i, stat
	
	tmp = 0.0
	if(adj) then
		do i=1,3
			m0 = (i-1)*n1*n2*n3 + 1
			m1 = i*n1*n2*n3
			call scaled_identity_init( eps(i) )
			stat = scaled_identity_exec( T, F, tmp(m0:m1), data(m0:m1) )
		end do
		stat = igrad3_lop( T, add, model, tmp)
	else
		stat = igrad3_lop( F, F, model, tmp)
		do i=1,3
			m0 = (i-1)*n1*n2*n3 + 1
			m1 = i*n1*n2*n3
			call scaled_identity_init( eps(i) )
			stat = scaled_identity_exec( F, add, tmp(m0:m1), data(m0:m1) )
		end do
	end if

	igrad3_weighted_exec = 0

end function igrad3_weighted_exec
!--------------------------------------------------------------------------------

!--------------------------------------------------------------------------------
  subroutine igrad3_close()
		deallocate( tmp)
  end subroutine 
!--------------------------------------------------------------------------------


end module 
