!--------------------------------------------------------------------------------
module lint3  
  use adj_mod
	use dottest
	use chain0_mod
	use mult_nmo_util

  implicit none

  integer, private, save :: m1, m2, m3
  real,    private, save :: o1, d1, o2, d2, o3, d3, wtpow

  real, private, dimension(:), pointer, save   :: wtlint, tmplint
  real, private, dimension(:,:), pointer, save :: xyz
	logical, private, save :: verbose, normlz
	logical, parameter, private, save :: T=.true., F=.false.

  contains

!--------------------------------------------------------------------------------
subroutine lint3_test()
	real :: num
	integer :: n1, n2, n3, nd, id, nlev
	real, dimension(:,:), pointer :: xyz_test

	n1=40;  n2=20;  n3=1;  nd = n1*n2*n3/10

	allocate( xyz_test(nd,3) )

	do id=1,nd
		call random_number(num)
		xyz_test(id,1) = (nint(num*n1)-1)*1.0
		call random_number(num)
		xyz_test(id,2) = (nint(num*n2)-1)*1.0
		xyz_test(id,3) = 0.0
	end do

	call lint3_init( (/n1,n2,n3/), (/0.0,0.0,0.0/), (/1.0,1.0,1.0/), xyz_test,    &
	                 norm=T                                                       )
	call lint3_dottest()
	call lint3_close()
	deallocate( xyz_test)
end subroutine lint3_test
!--------------------------------------------------------------------------------

!--------------------------------------------------------------------------------
subroutine lint3_init( m, o, d, xyz_in, norm, verb, foldpow )

	integer :: m(3), id
	real    :: o(3), d(3), foldpow
	real, dimension (:,:) :: xyz_in 
	logical  :: norm, verb
	optional :: norm, verb, foldpow

	verbose = .false.;  if(present(verb)) verbose=verb
	normlz  = .false.;  if(present(norm)) normlz=norm
	wtpow   = 1.0;      if(present(foldpow)) wtpow=foldpow
	m1 = m(1);  m2 = m(2);  m3 = m(3)
	o1 = o(1);  o2 = o(2);  o3 = o(3)
	d1 = d(1);  d2 = d(2);  d3 = d(3)
	allocate( xyz(size(xyz_in,dim=1),3) )
	xyz = xyz_in
	if(normlz) then
		allocate( wtlint(m1*m2*m3), tmplint(m1*m2*m3) )
		call lint3_weight_init( wtlint )
	end if
end subroutine lint3_init
!--------------------------------------------------------------------------------

!--------------------------------------------------------------------------------
subroutine lint3_weight_init( wt )
	real, dimension(:)  :: wt
	integer :: id, nd, im, imlo(3), imhi(3), im1d(8)
	real    :: wtlo(3), wthi(3), wt1d(8)

	nd = size(xyz,dim=1)
	wt = 0.0
	do id=1,nd
		call get_interpn_weights( xyz(id,:), (/d1,d2,d3/), (/o1,o2,o3/), imlo,    &
		                          imhi, wtlo, wthi, (/1,1,1/), (/m1,m2,m3/)       )
		call bins_weights_to_1d( (/m1,m2,m3/), wtlo, wthi, imlo, imhi, wt1d, im1d )

		do im=1,8
			wt(im1d(im)) = wt(im1d(im)) + wt1d(im)
		end do
	end do

	wt = wt**wtpow
end subroutine lint3_weight_init
!--------------------------------------------------------------------------------

!--------------------------------------------------------------------------------
subroutine get_lint3_interp_wt( wt )
	real, dimension(:)  :: wt
	if( .NOT. normlz) call seperr("get_lint3_interp_wt(): must have normlz=T")
	wt = wtlint
end subroutine get_lint3_interp_wt
!--------------------------------------------------------------------------------

!--------------------------------------------------------------------------------
integer function lint3_core( adj, add, model, data )
	logical, intent(IN) :: adj,add 
	real, dimension(:)  :: data, model 
	integer :: id, im, imlo(3), imhi(3), im1d(8)
	real    :: wtlo(3), wthi(3), wt1d(8)

	call adjnull( adj, add, model, data )

	if(verbose) call oper_stats( "lint3_core", T, model, data, adj)

	do id=1,size(data)
		call get_interpn_weights( xyz(id,:), (/d1,d2,d3/), (/o1,o2,o3/), imlo,    &
			                          imhi, wtlo, wthi, (/1,1,1/), (/m1,m2,m3/)       )
		call bins_weights_to_1d( (/m1,m2,m3/), wtlo, wthi, imlo, imhi, wt1d, im1d )
		do im=1,8
			if ( adj) then
				model(im1d(im)) = model(im1d(im)) + data(id)*wt1d(im)
			else
				data(id) = data(id) + model(im1d(im))*wt1d(im)
			end if
		end do
	end do

	if(verbose) call oper_stats( "lint3_core", F, model, data, adj)

end function lint3_core
!--------------------------------------------------------------------------------

!--------------------------------------------------------------------------------
integer function lint3_wt( adj, add, model, data )
	logical, intent(IN) :: adj,add 
	real, dimension(:)  :: model, data 

	call adjnull( adj, add, model, data )

	if(verbose) call oper_stats( "lint3_wt", T, model, data, adj)

	if( adj) then
		where( wtlint > 1e-3 )  model = model + data/wtlint
	else
		where( wtlint > 1e-3 )  data = data + model/wtlint
	end if

	if(verbose) call oper_stats( "lint3_wt", F, model, data, adj)
end function lint3_wt
!--------------------------------------------------------------------------------

!--------------------------------------------------------------------------------
integer function lint3_exec( adj, add, model, data )
	logical, intent(IN) :: adj,add 
	real, dimension(:)  :: model, data 
	integer             :: stat

	call adjnull( adj, add, model, data )

	if(verbose) call oper_stats( "lint3_exec", T, model, data, adj)

	if( normlz) then
		call chain0( lint3_core, lint3_wt, adj, add, model, data, tmplint)
	else
		stat = lint3_core( adj, add, model, data)
	end if

	if(verbose) call oper_stats( "lint3_exec", F, model, data, adj)
end function lint3_exec
!--------------------------------------------------------------------------------

!--------------------------------------------------------------------------------
subroutine lint3_close()
	if( normlz) then
		deallocate( xyz, wtlint, tmplint)
	else
		deallocate( xyz)
	end if
end subroutine lint3_close
!--------------------------------------------------------------------------------

!--------------------------------------------------------------------------------
subroutine get_lint3_fold(fold)
	real, dimension(:) :: fold, d_ones
	integer            :: stat

	allocate( d_ones(size(xyz,dim=1)) )
	d_ones = 1.0

	stat = lint3_core( T, F, fold, d_ones)
	deallocate( d_ones)
end subroutine get_lint3_fold
!--------------------------------------------------------------------------------

!--------------------------------------------------------------------------------
subroutine lint3_dottest()
	real :: dot1(2), dot2(2)
	write(0,*) '========= Dot Product Testing lint3_exec() ================'
	call dot_test( lint3_exec, m1*m2*m3, size(xyz,dim=1), dot1, dot2 )
	write(0,*) '========= Dot Product Testing lint3_core() ================'
	call dot_test( lint3_core, m1*m2*m3, size(xyz,dim=1), dot1, dot2 )
	write(0,*) '========= Dot Product Testing lint3_wt() =================='
	call dot_test( lint3_wt, m1*m2*m3, m1*m2*m3, dot1, dot2 )
	write(0,*) '==========================================================='
end subroutine lint3_dottest
!--------------------------------------------------------------------------------

end module lint3
