!-------------------------------------------------------------------------------
module ndipfilt3

  use sep 
  use nhelix
  use createnhelixmod
	use nhelicon
	use adj_mod
	use morgan_util
	use npolydiv

  implicit none

	type(nfilter), private, save :: nfilt2, nfilt3
	integer, private, save :: n1, n2, n3
	real, dimension(:), allocatable, private, save :: mtmp
	logical, private, save            :: verbose
	logical, private, save, parameter :: T=.true., F=.false.

	contains

!-------------------------------------------------------------------------------
! 2-D steering filters in a 3-D Space
!-------------------------------------------------------------------------------
function get_ndipfilt3( dip, n, d, nf, direc, mask) result(aa)

	integer            :: nf, direc, n(3), a(3), center(3), gap(3), n123, nf2,   &
	                      ip, i, k, iflt
	real               :: d(3), p
	real, dimension(:) :: dip, mask
	type(nfilter)      :: aa
	optional           :: mask
  real, dimension(:), allocatable :: flt, masktmp, pp
	
  n123 = product(n)

	nf2 = floor(nf/2.0)

	if(direc==2) then
		dip = -dip*d(2)/d(1)                               ! Convert to dimensionless
	else
		dip = -dip*d(3)/d(1)                               ! Convert to dimensionless
	end if

	where( dip >  1.0*nf2 )  dip =  1.0*nf2             ! No aliasing allowed
	where( dip < -1.0*nf2 )  dip = -1.0*nf2

	allocate( flt(nf) )

	center = (/nf2+1,1,1/)
	if(direc==2) then                                 ! Filter acts in 2-direction
		a = (/nf,2,1/)
		gap = (/0,1,0/)
	else                                              ! Filter acts in 3-direction
		a = (/nf,1,2/)
		gap = (/0,0,1/)
	end if

  allocate( pp(n123), masktmp(n123) ) 
  pp = (/ (ip, ip = 1, n123) /)

  aa = createnhelix( n, center, gap, a, int(pp) )         ! allocate PEF

	if( present(mask) ) then
		masktmp = mask
	else
		masktmp = 1.0
	end if

	do i=1,n123
		p = dip(i)
		flt=1.0
		do k=0,nf-1
			do iflt=0,nf-1
				if( k/=iflt) flt(k+1) = flt(k+1)*(p-nf2+iflt)/(k-iflt)
			end do
		end do
		aa%hlx(i)%flt = -flt*masktmp(i)
	end do
	
	deallocate( flt, masktmp, pp)

end function get_ndipfilt3
!-------------------------------------------------------------------------------

!-------------------------------------------------------------------------------
subroutine ndipfilt3_init( filt2_in, filt3_in, n, verb)
	type(nfilter) :: filt2_in, filt3_in
	integer :: n(3)
	logical  :: verb
	optional :: verb

	nfilt2 = filt2_in
	nfilt3 = filt3_in
	n1 = n(1);  n2 = n(2);  n3 = n(3)
	allocate( mtmp(product(n)) )
	verbose=F;  if(present(verb)) verbose=verb
end subroutine ndipfilt3_init
!-------------------------------------------------------------------------------

!-------------------------------------------------------------------------------
subroutine ndipfilt3_close( )
	deallocate( mtmp)
end subroutine ndipfilt3_close
!-------------------------------------------------------------------------------

!-------------------------------------------------------------------------------
integer function ndipfilt23_exec( adj, add, model, data)
	logical :: adj, add
	real, dimension(:) :: model, data
	integer :: stat, j0, j1
	
	call adjnull( adj, add, model, data)

	if(verbose) call oper_stats( "ndipfilt23_exec", T, model, data, adj)

	j0 = 1
	j1 = n1*n2*n3
	call nhelicon_init(nfilt2)
	if( adj) then
		stat = nhelicon_lop( T, F, mtmp, data(j0:j1) )
	else
		stat = nhelicon_lop( F, add, model, data(j0:j1) )
	end if
	j0 = n1*n2*n3+1
	j1 = 2*n1*n2*n3
	call nhelicon_init(nfilt3)
	if( adj) then
		stat = nhelicon_lop( T, T, mtmp, data(j0:j1) )
		model = model + mtmp
	else
		stat = nhelicon_lop( F, add, model, data(j0:j1) )
	end if

	if(verbose) call oper_stats( "ndipfilt23_exec", F, model, data, adj)

end function ndipfilt23_exec
!-------------------------------------------------------------------------------

!-------------------------------------------------------------------------------
subroutine ndipfilt3_test( aa, n, mm, direc)
	integer        :: n(3), direc, i1, i2, i3, j, stat
	type( nfilter) :: aa
	real, dimension(:) :: mm
	real, dimension(:), allocatable :: tmp

	allocate( tmp(product(n)) )

	do i3=1,n(3)
		do i2=1,n(2)
			do i1=1,n(1)
				j = grid2line3( n, (/i1,i2,i3/) )
				if( direc==2 ) then
					if( mod(i1,10)==0 .AND. mod(i2,10)==0 ) tmp(j) = 1.0
				else
					if( mod(i1,10)==0 .AND. mod(i3,5)==0 ) tmp(j) = 1.0
				end if
			end do
		end do
	end do

	call npolydiv_init( product(n), aa )
	stat = npolydiv_lop( F, F, tmp, mm)

	deallocate( tmp)

end subroutine ndipfilt3_test
!-------------------------------------------------------------------------------

end module ndipfilt3
