#include <stdio.h>
#include <math.h>
#include <float.h>

#include "dip3.h"
#include "divide1.h"
#include "allp3.h"
#include "bool.h"
#include "seperr.h"

static float ***u1, ***u2, ***dp;
static int n1, n2, n3, n, sign;

void dip3_init(int nx, int ny, int nz, float eps, float lam, int isign)
{
    n1=nx; n2=ny; n3=nz; n=n1*n2*n3;
    u1 = (float***) sepalloc3(n1,n2,n3,sizeof(float));
    u2 = (float***) sepalloc3(n1,n2,n3,sizeof(float));
    dp = (float***) sepalloc3(n1,n2,n3,sizeof(float));

    divide1_init (n1,eps,lam);
    sign = isign;
}

void dip3_close(void)
{
    free3 (n3,n2, (void***) u1);
    free3 (n3,n2, (void***) u2);
    free3 (n3,n2, (void***) dp);
    divide1_close();
}

void dip3(int dip, int niter, int nw, int nj, int verb, 
	     float ***u, float*** p)
{
    int i1, i2, i3, iter;
    float mean, usum, psum, ui, dpi, pi;
    allpass ap;
 
    ap = allpass_init (nw,nj,n1,n2,n3,p);

    for (iter =0; iter < niter; iter++) {
	if (dip == 1) {
	    allpass1 (0, ap, u,u2);
	    allpass1 (1, ap, u,u1);
	} else {
	    allpass2 (0, ap, u,u2);
	    allpass2 (1, ap, u,u1);
	}

	mean = 0.;
	for(i3=0; i3 < n3; i3++) {
	    for(i2=0; i2 < n2; i2++) {
		for(i1=0; i1 < n1; i1++) {
		    ui = u1[i3][i2][i1];
		    mean += ui*ui;
		}
	    }
	}
	if (mean == 0.) return;

	mean = sqrt (mean/n);

	usum = 0.;
	psum = 0.;

	for(i3=0; i3 < n3; i3++) {
	    for(i2=0; i2 < n2; i2++) {
		for(i1=0; i1 < n1; i1++) {
		    u1[i3][i2][i1] /= mean;
		    u2[i3][i2][i1] /= mean;
		    if (verb) {
			usum += u2[i3][i2][i1]*u2[i3][i2][i1];
			psum += p[i3][i2][i1];
		    }
		}
	    }
	}

	if (verb) fprintf(stderr, "%d %g %g\n", iter+1, sqrt(usum/n), psum/n);

	divide3 (n3, n2, u2, u1, dp);

	for(i3=0; i3 < n3; i3++) {
	    for(i2=0; i2 < n2; i2++) {
		for(i1=0; i1 < n1; i1++) {
		    dpi = dp[i3][i2][i1];
		    pi = p[i3][i2][i1];
		    if (sign && 
			fabs(pi) > FLT_EPSILON && 
			fabs(pi) > fabs(dpi)) {
			p[i3][i2][i1] *= exp(dpi/pi);
		    } else {			
			p[i3][i2][i1] += dpi;
		    }
		}
	    }
	}
    }
}
